from __future__ import annotations

import ast
import asyncio
import json
import re
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import wraps
from typing import Any, cast, overload

from .config import config
from .console import Console
from .default_tools import final_answer
from .llm import llm
from .mcp import AppConfig, McpTool, load_tools
from .schema import Messages, P, T
from .tools import function_schema


@overload
def react(
    model: str = "",
    tools: list[Callable] | None = None,
    mcp_config: str = "",
    max_steps: int = -1,
    *,
    response_format: None = None,
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str]]: ...


@overload
def react(
    model: str = "",
    tools: list[Callable] | None = None,
    mcp_config: str = "",
    max_steps: int = -1,
    *,
    response_format: type[T],
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, T]]: ...


def react(
    model: str = "",
    tools: list[Callable] | None = None,
    mcp_config: str = "",
    max_steps: int = -1,
    *,
    response_format: type[T] | None = None,
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str | T]]:
    default_model_from_decorator = model if model else config.default_model
    default_tools: list[Callable] = [] if tools is None else tools
    default_mcp_config = AppConfig.load(mcp_config)
    default_response_format = response_format  # noqa: F841
    default_api_params_from_decorator = api_params.copy()
    default_console = console or Console(show_message=False, show_react=True)

    def parameterized_lm_decorator(
        prompt: Callable[P, str | Messages],
    ) -> Callable[P, str | T]:
        @wraps(prompt)
        def model_call(
            *prompt_args: P.args,
            **prompt_kwargs: P.kwargs,
        ) -> str | T:
            default_console.init()
            default_console.show_result = False
            # Add final answer to the list of tools
            default_tools.insert(0, final_answer)

            def final_call(*prompt_args: P.args, acts: list[Action], **prompt_kwargs: P.kwargs):  # type: ignore
                output = prompt(*prompt_args, **prompt_kwargs)
                history = "\n".join(str(a) for a in acts)
                if isinstance(output, list):
                    output.append(
                        {"role": "assistant", "content": str(acts[-1]) + "\n-----\n Now you must give an final answer!"}
                    )
                    return output
                elif isinstance(output, str):
                    return output + f"\n-----{history}\n-----\n Now you must give an final answer!"
                else:
                    raise ValueError("Invalid output type")

            llm_final_call = llm(
                default_model_from_decorator,
                response_format=default_response_format,
                map_keys=None,
                n=None,
                tools=None,
                **default_api_params_from_decorator,
            )(final_call)

            def react_once(*prompt_args: P.args, acts: list[Action] | None = None, **prompt_kwargs: P.kwargs):  # type: ignore
                if acts is None:
                    acts = []
                output = prompt(*prompt_args, **prompt_kwargs)
                if isinstance(output, list):
                    if acts:
                        output.append({"role": "assistant", "content": str(acts[-1])})
                    return output
                elif isinstance(output, str):
                    return output + "\n".join(str(a) for a in acts)
                else:
                    raise ValueError("Invalid output type")

            default_console.log_model_usage_pre(default_model_from_decorator, prompt, prompt_args, prompt_kwargs)
            default_console.off()
            loop = asyncio.get_event_loop()
            with ThreadPoolExecutor() as executor:
                future = executor.submit(loop.run_until_complete, load_tools(default_mcp_config))
                mcp_clients, mcp_tools = future.result()

                tool_names = [f"`{tool.__name__}`" for tool in default_tools]
                tool_names.extend(f"`{tool.name}`" for tool in mcp_tools)
                react_once.__doc__ = (prompt.__doc__ if prompt.__doc__ else "") + SYSTEM_PROMPT.format(
                    tool_names=", ".join(tool_names),
                    tool_descriptions=_tool_descriptions(default_tools) + "\n" + _mcp_tool_descriptions(mcp_tools),
                )

                llm_tool_call = llm(
                    model=default_model_from_decorator,
                    console=default_console,
                    map_keys=None,
                    response_format=None,
                    n=None,
                    tools=None,
                    stop=["Observation:"],
                    **default_api_params_from_decorator,
                )(react_once)

                def _call_tool(name: str, args: dict[str, Any]) -> str:
                    if config.need_confirm:
                        if not default_console.confirm(f"Do you want to use tool {name}?"):
                            return "User cancelled. Please find other ways to solve this problem."
                    for tool in default_tools:
                        if tool.__name__ == name:
                            return tool(**args)
                    for mcp_tool in mcp_tools:
                        if mcp_tool.name == name:
                            future = executor.submit(loop.run_until_complete, mcp_tool._arun(**args))
                            return future.result()
                    raise ValueError(f"Can't find tool {name}")

                react_times = 0
                result = llm_tool_call(*prompt_args, acts=None, **prompt_kwargs)
                act = Action.from_response(result, _call_tool)
                default_console.log_react(act.__dict__)
                acts = [act]
                while not act.done and (max_steps == -1 or react_times < max_steps):
                    result = llm_tool_call(*prompt_args, acts=acts, **prompt_kwargs)
                    act = Action.from_response(result, _call_tool)
                    default_console.log_react(act.__dict__)
                    react_times += 1
                    acts.append(act)
                # Close all clients
                for client in mcp_clients:
                    executor.submit(loop.run_until_complete, client.close())
            loop.close()
            response: str | T = act.obs
            if not act.done and react_times >= max_steps or default_response_format is not None:
                response = llm_final_call(*prompt_args, acts=acts, **prompt_kwargs)
            default_console.stop()
            return response

        model_call.__api_params__ = default_api_params_from_decorator  # type: ignore
        model_call.__func__ = prompt  # type: ignore

        return model_call

    return parameterized_lm_decorator


SYSTEM_PROMPT = """You are an expert assistant who can solve any task using tools. You will be given a task to solve as best you can.
To do so, you have been given access to the following tools: [{tool_names}].

To solve the task, you must plan forward to proceed in a series of steps, in a cycle of 'Thought:', 'Action:', 'Action Input:', and 'Observation:' sequences.

At each step, in the 'Thought:' sequence, you should first explain your reasoning towards solving the task and the tools that you want to use.
Then in the 'Action:' and 'Action Input:' sequence, you should tell system what tool to use and what arguments to use.
The action result will then appear in the 'Observation:' field, which will be available as input for the next step. And you do not need to generate this part, it will be automatically filled by the system. The observation will always be a string: it can represent a file, like "image_1.jpg".
In the end you have to return a final answer using the `final_answer` tool.

## Example
An fake example:
```
Task:
the input question you must answer

Thought: you should always think about what to do
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action with JSON format representing the kwargs (e.g. {{"text": "hello world", "num_beams": 5}})
Observation: the result of the action

... (this Thought/Action/Action Input/Observation can be repeated zero or more times)

Thought: I now know the final answer
Action: final_answer
Action Input: {{"answer":"<answer>"}}
```

## Tools
{tool_descriptions}

## Instructions
1. ALWAYS provide a tool call, else you will fail.
2. Always use the right arguments for the tools. Never use variable names as the action arguments, use the value instead.
3. Call a tool only when needed: do not call the search agent if you do not need information, try to solve the task yourself. If no tool call is needed, use final_answer tool to return your answer.
4. Never re-do a tool call that you previously did with the exact same parameters.
"""


@dataclass
class Action:
    thought: str
    tool: str
    args: dict
    obs: str

    @property
    def done(self) -> bool:
        if self.tool == "final_answer":
            return True
        else:
            return False

    def __repr__(self) -> str:
        return self.info

    @property
    def info(self) -> str:
        if self.done:
            return f"\nThought: {self.thought}\nAction: Finish\nObservation: {self.obs}"
        return f"\nThought: {self.thought}\nAction: {self.tool}\nAction Input: {self.args}\nObservation: {self.obs}"

    @classmethod
    def from_response(cls, text: str, call: Callable[[str, dict[str, Any]], str]) -> Action:
        special_func_token = "\nAction:"
        special_args_token = "\nAction Input:"
        special_obs_token = "\nObservation:"
        func_name, func_args = None, None
        i = text.rfind(special_func_token)
        j = text.rfind(special_args_token)
        k = text.rfind(special_obs_token)
        if 0 <= i < j:  # If the text has `Action` and `Action input`,
            if k < j:  # but does not contain `Observation`,
                # then it is likely that `Observation` is omitted by the LLM,
                # because the output text may have discarded the stop word.
                text = text.rstrip() + special_obs_token  # Add it back.
            k = text.rfind(special_obs_token)
            func_name = text[i + len(special_func_token) : j].strip().split("#")[0]
            func_args = text[j + len(special_args_token) : k].strip().split("#")[0]
            match = re.search(r"(\{.*\})", func_args, re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if match:
                func_args = match.group()
            else:
                raise ValueError("Can't parse the Action Input")
            text = text[:i].strip()  # Return the response before tool call, i.e., `Thought`
            if text.startswith("Thought:"):
                text = text[len("Thought:") :]
        # return (func_name is not None), func_name, func_args, text
        if func_name is None or func_args is None:
            raise ValueError("Can't parse the response")
        tool = func_name
        args = ast.literal_eval(func_args)
        return cls(
            thought=text,
            tool=tool,
            args=args,
            obs=call(tool, args),  # type: ignore
        )


def _tool_descriptions(tools: list[Callable]) -> str:
    descriptions = []

    for tool in tools:
        schema = function_schema(tool)
        markdown = f"### {schema['name']}\n"
        markdown += f"{json.dumps(schema, ensure_ascii=False)}\n"
        descriptions.append(markdown)
    return "\n".join(descriptions)


def _mcp_tool_descriptions(mcp_tools: list[McpTool]) -> str:
    descriptions = []

    for tool in mcp_tools:
        schema = tool.args_schema
        schema.update({"name": tool.name})
        markdown = f"### {tool.name}\n"
        markdown += f"{json.dumps(schema, ensure_ascii=False)}\n"
        descriptions.append(markdown)
    return "\n".join(descriptions)



from __future__ import annotations

import os
import sys
from io import StringIO
from pathlib import Path

from uglychain.client import Client
from uglychain.default_tools import chat, execute_command, final_answer, user_input


def load_env_file(filepath):
    path = Path(filepath)
    with path.open() as f:
        for line in f:
            # 去除空行和注释
            if line.strip() and not line.startswith("#"):
                # 拆分键和值
                key, value = line.strip().split("=", 1)
                # 设置环境变量
                os.environ[key] = value
                print(f"{key} = {value}")


def test_final_answer():
    assert final_answer("Test Answer") == "Test Answer"


def test_user_input():
    input_str = "Test Input"
    sys.stdin = StringIO(input_str)
    assert user_input("Test Question") == input_str
    sys.stdin = sys.__stdin__


"""
def test_chat():
    Client.reset()
    # 加载 .env 文件
    load_env_file(".env")
    assert chat("Test Question")
"""


def test_execute_command():
    result = execute_command("echo Hello World")
    assert "Hello World" in result



from __future__ import annotations

import pytest
from rich.console import Console as RichConsole
from rich.live import Live
from rich.progress import Progress
from rich.table import Table, box

from uglychain.console import REACT_NAME, REACT_STYLE, Console


@pytest.fixture
def mock_config(monkeypatch):
    class MockConfig:
        verbose = True

    mock = MockConfig()
    monkeypatch.setattr("uglychain.console.config", mock)
    return mock


@pytest.fixture
def console(mock_config):
    return Console()


def test_console_initialization(console):
    assert isinstance(console.console, RichConsole)
    assert isinstance(console.progress, Progress)
    assert isinstance(console.react_table, Table)
    assert console.react_table.box == box.SIMPLE
    assert console.react_table.expand is True
    assert len(console.react_table.columns) == 2


def test_init_method(console, mock_config):
    console.init()
    assert console.show_base_info == (console.show_base_info and mock_config.verbose)
    assert console.show_progress == (console.show_progress and not mock_config.verbose)
    assert console.progress.disable == (not console.show_progress)


def test_log_model_usage_pre(console, mock_config):
    mock_config.verbose = False
    console.log_model_usage_pre("model", lambda: None, (), {})
    # Should not raise any errors


def test_log_progress_methods(console):
    console.log_progress_start(10)
    assert hasattr(console, "_task_id")

    console.log_progress_intermediate()
    console.log_progress_end()


def test_log_messages(console, mock_config):
    messages = [
        {"role": "system", "content": "test system"},
        {"role": "user", "content": "test user"},
        {"role": "assistant", "content": "test assistant"},
    ]
    console.log_messages(messages)
    # Should not raise any errors


def test_log_api_params(console, mock_config):
    api_params = {"tools": [{"function": {"name": "test_tool", "parameters": {"param1": "value1"}}}]}
    console.log_api_params(api_params)
    # Should not raise any errors


def test_log_results(console, mock_config):
    results = ["result1", "result2"]
    console.log_results(results)
    # Should not raise any errors


def test_log_react(console, mock_config):
    act = {"thought": "test thought", "tool": "test tool", "args": "test args", "obs": "test obs"}
    console.show_react = True
    console.log_react(act)
    assert len(console.react_table.rows) == 4


def test_off_method(console):
    console.off()
    assert console.show_base_info is False
    assert console.show_progress is False
    assert console.show_api_params is False
    assert console.show_result is False
    assert console.progress.disable is True



from __future__ import annotations

import inspect
import json
import re
from collections.abc import Callable
from enum import Enum, unique
from typing import Any, Generic, get_origin, get_type_hints

from openai.lib import _pydantic
from pydantic import BaseModel, ValidationError

from .schema import Messages, T, ToolResopnse


@unique
class Mode(Enum):
    TOOLS = "tool_call"
    MD_JSON = "markdown_json_mode"
    JSON_SCHEMA = "json_schema_mode"


class ResponseModel(Generic[T]):
    # 定义输出格式的提示模板，包含对JSON schema的描述和示例
    PROMPT_TEMPLATE = """## Output Format
The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{output_schema}
```

Make sure to return an instance of the JSON which can be parsed by Python json.loads, not the schema itself."""

    def __init__(self, func: Callable, response_model: type[T] | None = None) -> None:
        # 获取被修饰函数的返回类型
        response_type = get_type_hints(func).get("return", str) if response_model is None else response_model
        self.response_type: type[str] | type[T] = str if get_origin(response_type) is list else response_type
        self.mode: Mode = Mode.MD_JSON
        self._validate_response_type()

    def _validate_response_type(self) -> None:
        if self.response_type is not str and not issubclass(self.response_type, BaseModel):
            raise TypeError(f"Unsupported return type: {self.response_type}")

    def process_parameters(
        self,
        model: str,
        messages: Messages,
        merged_api_params: dict[str, Any],
    ) -> None:
        if self.response_type is str:
            return
        provider, model_name = model.split(":", 1)
        # 使用字典映射来替代 match-case 语句，提高可维护性
        mode_mapping = {
            ("openai", ""): Mode.JSON_SCHEMA,
            ("openai", "yi-large"): Mode.MD_JSON,
            ("openai", "*"): Mode.TOOLS,
            ("openrouter", "openai/gpt-4o"): Mode.JSON_SCHEMA,
            ("openrouter", "openai/gpt-4o-mini"): Mode.JSON_SCHEMA,
            ("openrouter", "*"): Mode.TOOLS,
            ("ollama", "*"): Mode.JSON_SCHEMA,
            ("gemini", "*"): Mode.JSON_SCHEMA,
        }
        # 使用通配符匹配
        self.mode = next(
            (mode for (p, m), mode in mode_mapping.items() if p == provider and (m == model_name or m == "*")),
            Mode.MD_JSON,
        )

        if self.mode == Mode.JSON_SCHEMA:
            self._set_response_format_from_params(merged_api_params)
        elif self.mode == Mode.TOOLS:
            self._set_tools_from_params(merged_api_params)
        elif self.mode == Mode.MD_JSON:
            self._update_markdown_json_schema_from_system_prompt(messages)

    def parse_from_response(self, choice: Any) -> str | T | ToolResopnse:
        # USE TOOLS
        if hasattr(choice.message, "tool_calls") and choice.message.tool_calls and self.mode != Mode.TOOLS:
            return ToolResopnse.parse(choice.message.tool_calls[0].function)
        # Other modes
        if self.response_type is str:
            return choice.message.content.strip()
        assert issubclass(self.response_type, BaseModel) and not inspect.isabstract(self.response_type)
        if self.mode == Mode.MD_JSON or self.mode == Mode.JSON_SCHEMA:
            response = choice.message.content.strip()
        elif self.mode == Mode.TOOLS:
            response = choice.message.tool_calls[0].function.arguments
        else:
            raise ValueError(f"Unsupported mode: {self.mode}")

        try:
            # 尝试直接解析整个响应字符串
            json_obj = json.loads(response.strip())
            return self.response_type.model_validate_json(json.dumps(json_obj))  # type: ignore[union-attr]

        except (json.JSONDecodeError, ValidationError):
            # 如果解析或验证失败，尝试使用正则表达式提取 JSON 字符串
            match = re.search(r"(\{.*\})", response.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
            if match:
                json_str = match.group()
                try:
                    return self.response_type.model_validate_json(json_str)  # type: ignore[union-attr]
                except (json.JSONDecodeError, ValidationError) as e:
                    # 如果解析或验证失败，记录错误信息并抛出异常
                    name = self.response_type.__name__
                    msg = f"Failed to parse {name} from completion {response}. Got: {e}"
                    raise ValueError(msg) from e
            else:
                # 如果正则表达式匹配失败，抛出异常
                name = self.response_type.__name__
                raise ValueError(f"Failed to find JSON object in response for {name}: {response}") from None

    def _update_markdown_json_schema_from_system_prompt(self, messages: Messages) -> None:
        if not messages:
            raise ValueError("Messages is empty")

        system_prompt = self.PROMPT_TEMPLATE.format(output_schema=json.dumps(self.parameters, ensure_ascii=False))
        system_message = messages[0]
        if system_message["role"] == "system":
            system_message["content"] += "\n\n" + system_prompt  # type: ignore
        else:
            messages.insert(
                0,
                {
                    "role": "system",
                    "content": system_prompt,
                },
            )

    def _set_response_format_from_params(self, api_params: dict[str, Any]) -> None:
        api_params["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "name": self.response_type.__name__,
                "schema": self.parameters,
                "strict": True,
            },
        }

    def _set_tools_from_params(self, api_params: dict[str, Any]) -> None:
        api_params["tools"] = [
            {
                "type": "function",
                "function": self.openai_schema,
            }
        ]
        api_params["tool_choice"] = {
            "type": "function",
            "function": {"name": self.openai_schema["name"]},
        }

    @property
    def schema(self) -> dict[str, Any]:
        if not hasattr(self, "_schema"):
            assert inspect.isclass(self.response_type) and issubclass(self.response_type, BaseModel)
            self._schema: dict[str, Any] = self.response_type.model_json_schema()
        return self._schema

    @property
    def parameters(self) -> dict[str, Any]:
        if not hasattr(self, "_parameters"):
            assert inspect.isclass(self.response_type) and issubclass(self.response_type, BaseModel)
            self._parameters: dict[str, Any] = _pydantic.to_strict_json_schema(self.response_type)
        return self._parameters

    @property
    def openai_schema(self) -> dict[str, Any]:
        if hasattr(self, "_openai_schema"):
            return self._openai_schema
        self._openai_schema: dict[str, Any] = {
            "name": self.schema["title"],
            "parameters": self.parameters,
            "strict": True,
        }
        if self.response_type.__doc__ is not None and self.response_type.__doc__.strip():
            self._openai_schema["description"] = self.response_type.__doc__
        else:
            self._openai_schema["description"] = (
                f"Correctly extracted `{self.schema['title']}` with all the required parameters with correct types"
            )

        return self._openai_schema



from __future__ import annotations

import asyncio
import json
import os
from dataclasses import dataclass, field
from typing import Any

import anyio
from mcp import ClientSession, StdioServerParameters, types
from mcp.client.stdio import stdio_client
from pydantic_core import to_json

DEFAULT_CONFIG = """{
  "mcpServers": {
  }
}"""


@dataclass
class ServerConfig:
    """Configuration for an MCP server."""

    command: str
    args: list[str] | None = None
    env: dict[str, str] | None = None
    enabled: bool = True
    exclude_tools: list[str] | None = None
    requires_confirmation: list[str] | None = None

    @classmethod
    def from_dict(cls, config: dict) -> ServerConfig:
        """Create ServerConfig from dictionary."""
        return cls(
            command=config["command"],
            args=config.get("args", []),
            env=config.get("env", {}),
            enabled=config.get("enabled", True),
            exclude_tools=config.get("exclude_tools", []),
            requires_confirmation=config.get("requires_confirmation", []),
        )


@dataclass
class AppConfig:
    mcp_servers: dict[str, ServerConfig]
    tools_requires_confirmation: list[str]

    @classmethod
    def load(cls, config_str: str = "") -> AppConfig:
        if not config_str:
            config_str = DEFAULT_CONFIG
        config = json.loads(config_str)

        # Extract tools requiring confirmation
        tools_requires_confirmation = []
        for server_config in config["mcpServers"].values():
            tools_requires_confirmation.extend(server_config.get("requires_confirmation", []))

        return cls(
            mcp_servers={
                name: ServerConfig.from_dict(server_config) for name, server_config in config["mcpServers"].items()
            },
            tools_requires_confirmation=tools_requires_confirmation,
        )

    def get_enabled_servers(self) -> dict[str, ServerConfig]:
        """Get only enabled server configurations."""
        return {name: config for name, config in self.mcp_servers.items() if config.enabled}


@dataclass
class McpServerConfig:
    server_name: str
    server_param: StdioServerParameters
    exclude_tools: list[str] = field(default_factory=list)


@dataclass
class McpTool:
    toolkit_name: str
    name: str
    description: str
    args_schema: dict[str, Any]
    session: ClientSession
    client: McpClient

    async def _arun(self, **kwargs: Any) -> str:
        if not self.session:
            self.session = await self.client._start_session()

        result = await self.session.call_tool(self.name, arguments=kwargs)  # type: ignore
        content = to_json(result.content).decode()
        return content


@dataclass
class McpClient:
    name: str
    server_param: StdioServerParameters
    exclude_tools: list[str] = field(default_factory=list)
    _session: ClientSession | None = None
    _tools: list[McpTool] = field(default_factory=list)
    _init_lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    async def _start_session(self) -> ClientSession:
        async with self._init_lock:
            if self._session is None:
                self._client = stdio_client(self.server_param)
                read, write = await self._client.__aenter__()
                self._session = ClientSession(read, write)
                await self._session.__aenter__()
                await self._session.initialize()
            return self._session

    async def initialize(self, force_refresh: bool = False) -> None:
        if self._tools and not force_refresh:
            return

        try:
            await self._start_session()
            tools: types.ListToolsResult = await self._session.list_tools()  # type: ignore
            for tool in tools.tools:
                if tool.name in self.exclude_tools:
                    continue
                self._tools.append(
                    McpTool(
                        self.name,
                        tool.name,
                        tool.description if tool.description else "",
                        tool.inputSchema,
                        self._session,  # type: ignore
                        self,
                    )
                )
        except Exception as e:
            print(f"Error gathering tools for {self.server_param.command} {' '.join(self.server_param.args)}: {e}")
            raise e

    async def close(self) -> None:
        try:
            if self._session:
                await self._session.__aexit__(None, None, None)
        except Exception:
            # Currently above code doesn't really works and not closing the session
            # But it's not a big deal as we are exiting anyway
            # TODO find a way to cleanly close the session
            pass
        try:
            if self._client:
                await self._client.__aexit__(None, None, None)
        except Exception:
            # TODO find a way to cleanly close the client
            pass

    @property
    def tools(self) -> list[McpTool]:
        return self._tools


async def load_tools(app_config: AppConfig, force_refresh: bool = False) -> tuple[list[McpClient], list[McpTool]]:
    server_configs = [
        McpServerConfig(
            server_name=name,
            server_param=StdioServerParameters(
                command=config.command, args=config.args or [], env={**(config.env or {}), **os.environ}
            ),
            exclude_tools=config.exclude_tools or [],
        )
        for name, config in app_config.get_enabled_servers().items()
    ]

    clients: list[McpClient] = []
    tools: list[McpTool] = []

    async def convert_toolkit(server_config: McpServerConfig) -> None:
        client = McpClient(server_config.server_name, server_config.server_param, server_config.exclude_tools)
        await client.initialize(force_refresh=force_refresh)
        clients.append(client)
        tools.extend(client.tools)

    async with anyio.create_task_group() as tg:
        for server_param in server_configs:
            tg.start_soon(convert_toolkit, server_param)

    return clients, tools



from __future__ import annotations

import inspect
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps
from typing import Any, overload

from .client import Client
from .config import config
from .console import Console
from .schema import Messages, P, T, ToolResopnse
from .structured import ResponseModel


@overload
def llm(
    func: Callable[P, str | Messages],
    /,
) -> Callable[P, str]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: None = None,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str | ToolResopnse]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: int,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str]]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: int,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str] | list[ToolResopnse]]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: type[T],
    console: Console | None = None,
    n: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, T]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: None = None,
    response_format: type[T],
    console: Console | None = None,
    n: int,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[T]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: None = None,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, str | ToolResopnse]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: type[T],
    console: Console | None = None,
    n: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, T]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: list[str],
    response_format: None = None,
    console: Console | None = None,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str]]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: list[str],
    response_format: None = None,
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str] | list[ToolResopnse]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: int,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: None = None,
    console: Console | None = None,
    n: int,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str] | list[ToolResopnse]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: list[str],
    response_format: None = None,
    console: Console | None = None,
    tools: None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: list[str],
    response_format: None = None,
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[str] | list[ToolResopnse]]]: ...
@overload
def llm(
    model: str,
    /,
    *,
    map_keys: list[str],
    response_format: type[T],
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[T]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: None = None,
    response_format: type[T],
    console: Console | None = None,
    n: int,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[T]]]: ...
@overload
def llm(
    *,
    model: str = "",
    map_keys: list[str],
    response_format: type[T],
    console: Console | None = None,
    **api_params: Any,
) -> Callable[[Callable[P, str | Messages]], Callable[P, list[T]]]: ...


def llm(
    func_or_model: Callable[P, str | Messages] | str | None = None,
    /,
    *,
    model: str = "",
    map_keys: list[str] | None = None,
    response_format: type[T] | None = None,
    console: Console | None = None,
    **api_params: Any,
) -> (
    Callable[P, str]
    | Callable[
        [Callable[P, str | Messages]], Callable[P, str | ToolResopnse | T | list[str] | list[ToolResopnse] | list[T]]
    ]
):
    """
    LLM 装饰器，用于指定语言模型和其参数。

    :param model: 模型名称
    :param api_params: API 参数，以关键字参数形式传入
    :return: 返回一个装饰器，用于装饰提示函数
    """
    if isinstance(func_or_model, str):
        model = func_or_model
        func = None
    elif callable(func_or_model):
        func = func_or_model
    else:
        func = None
    default_model_from_decorator = model if model else config.default_model
    default_console = console or Console()
    default_api_params_from_decorator = api_params.copy()

    def parameterized_lm_decorator(
        prompt: Callable[P, str | Messages],
    ) -> Callable[P, str | ToolResopnse | T | list[str] | list[ToolResopnse] | list[T]]:
        @wraps(prompt)
        def model_call(
            *prompt_args: P.args,
            api_params: dict[str, Any] | None = None,  # type: ignore
            **prompt_kwargs: P.kwargs,
        ) -> str | ToolResopnse | T | list[str] | list[ToolResopnse] | list[T]:
            default_console.init()
            # 获取被修饰函数的返回类型
            response_model = ResponseModel[T](prompt, response_format)

            # 合并装饰器级别的API参数和函数级别的API参数
            merged_api_params = config.default_api_params.copy()
            if "tools" in merged_api_params:
                merged_api_params.pop("tools")
            if default_api_params_from_decorator:
                merged_api_params.update(default_api_params_from_decorator)
            if api_params:
                if "tools" in api_params:
                    api_params.pop("tools")
                merged_api_params.update(api_params)

            # 获取同时运行的次数
            n = merged_api_params.get("n", 1)
            if n is None:
                n = 1
            # 获取模型名称
            model = merged_api_params.pop("model", default_model_from_decorator)

            default_console.log_model_usage_pre(model, prompt, prompt_args, prompt_kwargs)

            m, map_args_index_set, map_kwargs_keys_set = _get_map_keys(prompt, prompt_args, prompt_kwargs, map_keys)
            if m > 1 and n > 1:
                raise ValueError("n > 1 和列表长度 > 1 不能同时成立")
            if m > 1 or n > 1:
                default_console.show_result = False
            else:
                default_console.progress.disable = True

            def process_single_prompt(i: int) -> list[Any]:
                args = [arg[i] if j in map_args_index_set else arg for j, arg in enumerate(prompt_args)]  # type: ignore
                kwargs = {
                    key: value[i] if key in map_kwargs_keys_set else value  # type: ignore
                    for key, value in prompt_kwargs.items()
                }
                res = prompt(*args, **kwargs)  # type: ignore
                assert (
                    isinstance(res, str) or isinstance(res, list) and all(isinstance(item, dict) for item in res)
                ), ValueError("被修饰的函数返回值必须是 str 或 `messages`(list[dict[str, str]]) 类型")
                messages = _get_messages(res, prompt)
                response_model.process_parameters(model, messages, merged_api_params)

                default_console.log_api_params(merged_api_params)
                default_console.log_messages(messages)

                response = Client.generate(model, messages, **merged_api_params)

                # 从响应中解析结果
                result = [response_model.parse_from_response(choice) for choice in response]
                default_console.log_progress_intermediate()
                default_console.log_results(result)
                return result

            results = []
            default_console.log_progress_start(m if m > 1 else n)

            if config.use_parallel_processing:
                with ThreadPoolExecutor() as executor:
                    futures = [executor.submit(process_single_prompt, i) for i in range(m)]

                    for future in as_completed(futures):
                        results.extend(future.result())
            else:
                for i in range(m):
                    results.extend(process_single_prompt(i))

            default_console.log_progress_end()

            if len(results) == 0:
                raise ValueError("模型未返回任何选择")
            elif "n" in merged_api_params and merged_api_params["n"] is not None or map_keys:
                return results
            return results[0]

        model_call.__api_params__ = default_api_params_from_decorator  # type: ignore
        model_call.__func__ = prompt  # type: ignore

        return model_call  # type: ignore

    # 检查是否直接应用装饰器而不是传递参数
    if func is not None:
        return parameterized_lm_decorator(func)  # type: ignore

    return parameterized_lm_decorator


def _get_map_keys(
    prompt: Callable, prompt_args: tuple, prompt_kwargs: dict, map_keys: list[str] | None
) -> tuple[int, set[int], set[str]]:
    if map_keys is None:
        return 1, set(), set()
    map_key_set: set[str] = set(map_keys)
    map_num_set: set[int] = set()
    param_mapping = inspect.getcallargs(prompt, *prompt_args, **prompt_kwargs)

    list_lengths = []
    for i, (param_name, arg_value) in enumerate(param_mapping.items()):
        if param_name in map_key_set:
            if not isinstance(arg_value, list):
                raise ValueError("map_key 必须是列表")
            if i < len(prompt_args):
                map_num_set.add(i)
                map_key_set.remove(param_name)
            list_lengths.append(len(arg_value))

    if not list_lengths:
        return 1, set(), set()

    unique_lengths = set(list_lengths)
    if len(unique_lengths) > 1:
        raise ValueError("prompt_args 和 prompt_kwargs 中的 map_key 列表必须具有相同的长度")

    return list_lengths[0], map_num_set, map_key_set


def _get_messages(prompt_ret: str | Messages, prompt: Callable) -> Messages:
    if isinstance(prompt_ret, str):
        messages = []
        if prompt.__doc__ and prompt.__doc__.strip():
            messages.append({"role": "system", "content": prompt.__doc__})
        messages.append({"role": "user", "content": prompt_ret})
        return messages
    elif isinstance(prompt_ret, list):
        messages = prompt_ret
        if (not messages or messages[0]["role"] != "system") and prompt.__doc__ and prompt.__doc__.strip():
            messages.insert(0, {"role": "system", "content": prompt.__doc__})
        return messages
    else:
        raise TypeError("Expected prompt_ret to be a str or list of Messages")



from __future__ import annotations

import pytest

from uglychain.client import Client


def test_client_get(monkeypatch):
    class MockClient:
        def __init__(self):
            self.name = "MockClient"

    Client.reset()
    monkeypatch.setattr("aisuite.Client", MockClient)

    client1 = Client.get()
    client2 = Client.get()

    assert client1 is client2
    assert isinstance(client1, MockClient)


def test_client_reset(monkeypatch):
    class MockClient:
        def __init__(self):
            self.name = "MockClient"

    Client.reset()
    monkeypatch.setattr("aisuite.Client", MockClient)

    client1 = Client.get()
    Client.reset()
    client2 = Client.get()

    assert client1 is not client2
    assert isinstance(client2, MockClient)


def test_client_generate(monkeypatch):
    class MockClient:
        class chat:  # noqa: N801
            class completions:  # noqa: N801
                @staticmethod
                def create(model, messages, **kwargs):
                    return type(
                        "Response",
                        (object,),
                        {
                            "choices": [
                                type(
                                    "Choice",
                                    (object,),
                                    {"message": type("Message", (object,), {"content": "Test response"})},
                                )
                            ]
                        },
                    )

    Client.reset()
    monkeypatch.setattr("aisuite.Client", MockClient)

    response = Client.generate(
        model="test:model",
        messages=[{"role": "user", "content": "Hello"}],
    )

    assert response[0].message.content == "Test response"


def test_client_generate_exception(monkeypatch):
    class MockClient:
        class chat:  # noqa: N801
            class completions:  # noqa: N801
                @staticmethod
                def create(model, messages, **kwargs):
                    raise Exception("Test exception")

    Client.reset()
    monkeypatch.setattr("aisuite.Client", MockClient)

    with pytest.raises(RuntimeError) as exc_info:
        Client.generate(
            model="test:model",
            messages=[{"role": "user", "content": "Hello"}],
        )

    assert "生成响应失败: Test exception" in str(exc_info.value)


def test_client_generate_no_choices(monkeypatch):
    class MockClient:
        class chat:  # noqa: N801
            class completions:  # noqa: N801
                @staticmethod
                def create(model, messages, **kwargs):
                    return type("Response", (object,), {})

    Client.reset()
    monkeypatch.setattr("aisuite.Client", MockClient)

    with pytest.raises(ValueError) as exc_info:
        Client.generate(
            model="test:model",
            messages=[{"role": "user", "content": "Hello"}],
        )

    assert "No choices returned from the model" in str(exc_info.value)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field


class Config(BaseModel):
    default_model: str = Field(default="openai:gpt-4o-mini", description="默认语言模型。")
    default_api_params: dict[str, Any] = Field(default_factory=dict, description="语言模型的默认参数。")
    use_parallel_processing: bool = Field(default=False, description="是否对语言模型使用并行处理。")
    verbose: bool = Field(default=False, description="如果为真，则启用详细日志记录。")
    need_confirm: bool = Field(default=False, description="如果为真，则工具使用需要确认。")

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


config = Config()



from __future__ import annotations

import inspect
import json
from collections.abc import Callable
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any

import rich
from rich.columns import Columns
from rich.console import Group
from rich.live import Live
from rich.progress import Progress
from rich.prompt import Confirm, Prompt
from rich.table import Table, box

from .config import config
from .schema import Messages

MAX_AGRS: int = 5
MAX_ARGS_LEN: int = 8

REACT_NAME = {
    "thought": "Thought",
    "tool": "Action",
    "args": "Action Input",
    "obs": "Observation",
}

REACT_STYLE = {
    "thought": "italic yellow",
    "tool": "bold red",
    "args": "bold red",
    "obs": "bold green",
}


@dataclass
class Console:
    show_base_info: bool = True
    show_progress: bool = True
    show_api_params: bool = True
    show_result: bool = False
    show_message: bool = True
    show_react: bool = False
    console: rich.console.Console = field(init=False, default_factory=rich.console.Console)
    progress: Progress = field(init=False, default_factory=Progress)
    react_table: Table = field(init=False, default_factory=Table)
    messages_table: Table = field(init=False, default_factory=Table)
    _live: Live = field(init=False)

    def __post_init__(self) -> None:
        self._init_react_table()
        self._init_messages_table()

    @property
    def group(self) -> Group:
        _group = []
        if self.show_message:
            _group.append(self.messages_table)
        if self.show_react:
            _group.append(self.react_table)
        return Group(*_group)

    def init(self) -> None:
        self.show_base_info = self.show_base_info and config.verbose
        self.show_progress = self.show_progress and not config.verbose
        self.progress.disable = not self.show_progress

        self.show_message = self.show_message and config.verbose
        self.show_api_params = self.show_api_params and config.verbose
        self.show_result = self.show_result and config.verbose

    def _init_react_table(self) -> None:
        self.react_table = Table(box=box.SIMPLE, show_header=False, expand=True)
        self.react_table.add_column("Type")
        self.react_table.add_column("Information")

    def _init_messages_table(self) -> None:
        self.messages_table = Table(title="Prompt", box=box.SIMPLE, show_header=False, expand=True)
        self.messages_table.add_column("角色", justify="right", no_wrap=True)
        self.messages_table.add_column("内容")

    def log_model_usage_pre(
        self,
        model: str,
        func: Callable,
        args: tuple[object, ...],
        kwargs: dict[str, Any],
    ) -> None:
        if not self.show_base_info:
            return
        base_table = Table(box=box.SIMPLE, show_header=False)
        base_table.add_column(style="green")
        base_table.add_column(style="blue")
        base_table.add_row("Model", model)
        base_table.add_row("Func", _format_func_call(func, *args, **kwargs))
        self.console.print(base_table)

    def log_progress_start(self, n: int) -> None:
        self._task_id = self.progress.add_task("模型进度", total=n)
        self.progress.start()

    def log_progress_intermediate(self) -> None:
        self.progress.update(self._task_id, advance=1)

    def log_progress_end(self) -> None:
        self.progress.stop()

    def log_messages(self, messages: Messages) -> None:
        if not self.show_message:
            return
        self._init_messages_table()
        for message in messages:
            if message["role"] == "system":
                style = "bold green"
            elif message["role"] == "user":
                style = "italic yellow"
            else:
                style = "italic blue"
            self.messages_table.add_row(str(message["role"]), str(message["content"]), style=style)
        self._update_live()

    def log_api_params(
        self,
        api_params: dict[str, Any],
    ) -> None:
        if not self.show_api_params:
            return
        if api_params:
            if "tools" in api_params and api_params["tools"]:
                params_table = Table(title="Tools", box=box.SIMPLE, expand=True)
                params_table.add_column("Name", justify="right", no_wrap=True)
                params_table.add_column("Parameters", justify="center")
                for tool in api_params["tools"]:
                    params_table.add_row(
                        tool["function"]["name"], json.dumps(tool["function"]["parameters"], ensure_ascii=False)
                    )
                self.console.print(params_table)

    def log_results(self, result: list) -> None:
        if not self.show_result:
            return
        self.console.print(Columns([i.model_dump_json(indent=2) if not isinstance(i, str) else i for i in result]))

    def log_react(self, act: dict[str, Any]) -> None:
        if not self.show_react:
            return
        for key in ["thought", "tool", "args", "obs"]:
            self.react_table.add_row(REACT_NAME[key], str(act[key]).strip(), style=REACT_STYLE[key])
        self._update_live()

    def input(self, prompt: str, default: str = "") -> str:
        return Prompt.ask(prompt, console=self.console, default=default)

    def confirm(self, prompt: str, default: bool = True) -> bool:
        with PauseLive(self._get_live()):
            with self.console.status(prompt):
                output = Confirm.ask("", console=self.console, show_default=default)
        return output

    def off(self) -> None:
        self.show_base_info = False
        self.show_progress = False
        self.show_api_params = False
        self.show_result = False
        self.progress.disable = True

    def _get_live(self) -> Live:
        if not hasattr(self, "_live") or not self._live:
            self._live = Live(self.group, console=self.console, auto_refresh=False)
        if not self._live.is_started:
            self._live.start(refresh=True)
        return self._live

    def _update_live(self) -> None:
        live = self._get_live()
        live.update(self.group, refresh=True)

    def _stop_live(self) -> None:
        if hasattr(self, "_live") and self._live and self._live.is_started:
            self._live.stop()

    def stop(self) -> None:
        self._stop_live()


class PauseLive:
    def __init__(self, live: Live) -> None:
        self._live = live

    def __enter__(self) -> Live:
        self._live.stop()
        return self._live

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self._live.start()


def _format_arg_str(arg_str: Any) -> str:
    if isinstance(arg_str, str):
        if len(arg_str) > MAX_ARGS_LEN:
            return f"'{arg_str[:MAX_ARGS_LEN].strip()}...'"
        else:
            return f"'{arg_str}'"
    else:
        arg_str = repr(arg_str)
        if len(arg_str) > MAX_ARGS_LEN:
            return f"{arg_str[:MAX_ARGS_LEN].strip()}..."
        else:
            return arg_str


def _format_func_call(func: Callable, *args: Any, **kwargs: Any) -> str:
    # 获取函数的参数信息
    signature = inspect.signature(func)
    bound_arguments = signature.bind(*args, **kwargs)
    bound_arguments.apply_defaults()

    # 构建参数字符串
    args_str = []
    for name, value in bound_arguments.arguments.items():
        if isinstance(value, list | dict | set | tuple):
            end_str = ""
            if len(value) > 2:
                value = list(value)[:2]
                end_str = ",..."
            if isinstance(value, list):
                display_value = f"[{', '.join([_format_arg_str(i) for i in value[:2]])}{end_str}]"
            elif isinstance(value, dict):
                display_value = f"{{{', '.join([f'{_format_arg_str(k)}: {_format_arg_str(value[k])}' for k in list(value)[:2]])}{end_str}}}"
            elif isinstance(value, set):
                display_value = f"{{{', '.join([_format_arg_str(i) for i in list(value)[:2]])}{end_str}}}"
            elif isinstance(value, tuple):
                display_value = f"({', '.join([_format_arg_str(i) for i in value[:2]])}{end_str})"
            else:
                display_value = value
            args_str.append(f"{name}={display_value}")
        else:
            args_str.append(f"{name}={_format_arg_str(value)}")

    # 构建最终的函数调用字符串
    if len(args_str) < MAX_AGRS:
        func_call_str = f"{func.__name__}({', '.join(args_str)})"
    else:
        func_call_str = f"{func.__name__}({', '.join(args_str[:MAX_AGRS])}, ...)"
    return func_call_str



from __future__ import annotations

import shlex
import subprocess
from pathlib import Path

from .llm import llm


def final_answer(answer: str) -> str:
    """When get Final Answer, use this tool to return the answer and finishes the task."""
    return answer


def user_input(question: str) -> str:
    """
    Asks for user's input on a specific question
    """
    user_input = input(f"{question} => Type your answer here:")
    return user_input


@llm
def chat(question: str) -> str:
    """
    You are a helpful assistant. You are designed to answer questions for users.
    """
    return f"{question}"


chat.__doc__ = """
Get the answer to the question input from Large Language Models.
"""


def execute_command(command: str) -> str:
    """
    Request to execute a CLI command on the system. Use this when you need to perform system operations or run specific commands to accomplish any step in the user's task. You must tailor your command to the user's system and provide a clear explanation of what the command does. Prefer to execute complex CLI commands over creating executable scripts, as they are more flexible and easier to run. Commands will be executed in the specified working directory.
    """

    if not command:
        raise ValueError("Command cannot be empty")

    working_directory = Path.cwd()

    result = subprocess.run(shlex.split(command), cwd=working_directory, capture_output=True, text=True)
    return result.stdout + result.stderr


__all__ = [
    "final_answer",
    "execute_command",
    "user_input",
    "chat",
]



from __future__ import annotations

import json

import pytest

from uglychain.schema import ToolResopnse


def test_tool_response_parse():
    response = type("Response", (object,), {"name": "test_tool", "arguments": '{"arg1": "value1"}'})()
    tool_response = ToolResopnse.parse(response)
    assert tool_response.name == "test_tool"
    assert tool_response.parameters == {"arg1": "value1"}


def test_tool_response_parse_invalid():
    response = type("Response", (object,), {"name": "test_tool", "arguments": "invalid_json"})()
    with pytest.raises(json.JSONDecodeError):
        ToolResopnse.parse(response)


def test_tool_response_run_function():
    def test_tool(arg1):
        return f"arg1: {arg1}"

    tool_response = ToolResopnse(name="test_tool", parameters={"arg1": "value1"})
    result = tool_response.run_function([test_tool])
    assert result == "arg1: value1"


def test_tool_response_run_function_tool_not_found():
    tool_response = ToolResopnse(name="non_existent_tool", parameters={"arg1": "value1"})
    with pytest.raises(ValueError, match="Can't find tool non_existent_tool"):
        tool_response.run_function([])



from __future__ import annotations

import inspect
import textwrap
import warnings
from collections.abc import Callable, Sequence
from typing import Annotated, Any, get_args, get_origin

from openai.lib import _pydantic
from pydantic import BaseModel, ConfigDict, PydanticDeprecationWarning, create_model, validate_arguments
from pydantic.fields import FieldInfo


def function_schema(func: Callable) -> dict[str, Any]:
    pydantic_func = create_schema_from_function(func.__name__, func)
    return convert_pydantic_to_openai_function(pydantic_func)


def get_tools_schema(tools: list[Callable]) -> list[dict[str, Any]]:
    if not tools:
        return []
    return [{"type": "function", "function": function_schema(tool)} for tool in tools]


def add_tools_to_parameters(params: dict[str, Any], tools: list[Callable] | None) -> None:
    if tools is None or not tools:
        return
    params["tools"] = []
    for tool in tools:
        params["tools"].append({"type": "function", "function": function_schema(tool)})
    # TODO: 这里的逻辑后续需要重新确认
    if len(tools) == 1:
        params["tool_choice"] = {"type": "function", "function": {"name": tools[0].__name__}}


def convert_pydantic_to_openai_function(
    model: type,
    *,
    rm_titles: bool = True,
) -> dict[str, Any]:
    if hasattr(model, "model_json_schema"):
        parameters = _pydantic.to_strict_json_schema(model)
        if "$defs" in parameters:  # pydantic 2
            parameters.pop("$defs", None)
        title = parameters.pop("title", "")
        default_description = parameters.pop("description", "")
        return {
            "name": model.__name__ or title,
            "description": model.__doc__ or default_description,
            "parameters": _rm_titles(parameters) if rm_titles else parameters,
        }
    else:
        raise ValueError("Model must be a Pydantic 2 model.")


class _SchemaConfig:
    """Configuration for the pydantic model.

    This is used to configure the pydantic model created from
    a function's signature.

    Parameters:
        extra: Whether to allow extra fields in the model.
        arbitrary_types_allowed: Whether to allow arbitrary types in the model.
            Defaults to True.
    """

    extra: str = "forbid"
    arbitrary_types_allowed: bool = True


FILTERED_ARGS = ("run_manager", "callbacks")


def create_schema_from_function(
    model_name: str,
    func: Callable,
    *,
    filter_args: Sequence[str] | None = None,
    include_injected: bool = True,
) -> type[BaseModel]:
    """Create a pydantic schema from a function's signature.

    Args:
        model_name: Name to assign to the generated pydantic schema.
        func: Function to generate the schema from.
        filter_args: Optional list of arguments to exclude from the schema.
            Defaults to FILTERED_ARGS.
        parse_docstring: Whether to parse the function's docstring for descriptions
            for each argument. Defaults to False.
        error_on_invalid_docstring: if ``parse_docstring`` is provided, configure
            whether to raise ValueError on invalid Google Style docstrings.
            Defaults to False.
        include_injected: Whether to include injected arguments in the schema.
            Defaults to True, since we want to include them in the schema
            when *validating* tool inputs.

    Returns:
        A pydantic model with the same arguments as the function.
    """
    sig = inspect.signature(func)

    with warnings.catch_warnings():
        # We are using deprecated functionality here.
        # This code should be re-written to simply construct a pydantic model
        # using inspect.signature and create_model.
        warnings.simplefilter("ignore", category=PydanticDeprecationWarning)
        validated = validate_arguments(func, config=_SchemaConfig)  # type: ignore
    # Let's ignore `self` and `cls` arguments for class and instance methods
    # If qualified name has a ".", then it likely belongs in a class namespace
    in_class = bool(func.__qualname__ and "." in func.__qualname__)

    has_args = False
    has_kwargs = False

    for param in sig.parameters.values():
        if param.kind == param.VAR_POSITIONAL:
            has_args = True
        elif param.kind == param.VAR_KEYWORD:
            has_kwargs = True

    inferred_model = validated.model  # type: ignore

    if filter_args:
        filter_args_: list[str] = list(filter_args)
    else:
        # Handle classmethods and instance methods
        existing_params: list[str] = list(sig.parameters.keys())
        if existing_params and existing_params[0] in ("self", "cls") and in_class:
            filter_args_ = [existing_params[0]] + list(FILTERED_ARGS)
        else:
            filter_args_ = list(FILTERED_ARGS)

        for existing_param in existing_params:
            if not include_injected and _is_injected_arg_type(sig.parameters[existing_param].annotation):
                filter_args_.append(existing_param)

    description, arg_descriptions = _infer_arg_descriptions(func)
    # Pydantic adds placeholder virtual fields we need to strip
    valid_properties = []
    for field in inferred_model.model_fields:
        if not has_args and field == "args":
            continue
        if not has_kwargs and field == "kwargs":
            continue

        if field == "v__duplicate_kwargs":  # Internal pydantic field
            continue

        if field not in filter_args_:
            valid_properties.append(field)

    return _create_subset_model(
        model_name,
        inferred_model,
        list(valid_properties),
        descriptions=arg_descriptions,
        fn_description=description,
    )


class InjectedToolArg:
    """Annotation for a Tool arg that is **not** meant to be generated by a model."""


def _is_injected_arg_type(type_: type, injected_type: type[InjectedToolArg] | None = None) -> bool:
    injected_type = injected_type or InjectedToolArg
    return any(
        isinstance(arg, injected_type) or (isinstance(arg, type) and issubclass(arg, injected_type))
        for arg in get_args(type_)[1:]
    )


def _infer_arg_descriptions(
    fn: Callable,
) -> tuple[str, dict]:
    """Infer argument descriptions from a function's docstring."""
    if hasattr(inspect, "get_annotations"):
        # This is for python < 3.10
        annotations = inspect.get_annotations(fn)  # type: ignore
    else:
        annotations = getattr(fn, "__annotations__", {})

    description = inspect.getdoc(fn) or ""
    arg_descriptions = {}
    for arg, arg_type in annotations.items():
        if arg in arg_descriptions:
            continue
        if desc := _get_annotation_description(arg_type):
            arg_descriptions[arg] = desc
    return description, arg_descriptions


def _is_annotated_type(typ: type[Any]) -> bool:
    return get_origin(typ) is Annotated


def _get_annotation_description(arg_type: type) -> str | None:
    if _is_annotated_type(arg_type):
        annotated_args = get_args(arg_type)
        for annotation in annotated_args[1:]:
            if isinstance(annotation, str):
                return annotation
    return None


def _create_subset_model(
    name: str,
    model: type[BaseModel],
    field_names: list[str],
    *,
    descriptions: dict | None = None,
    fn_description: str | None = None,
) -> type[BaseModel]:
    descriptions_ = descriptions or {}
    fields = {}
    for field_name in field_names:
        field = model.model_fields[field_name]  # type: ignore
        description = descriptions_.get(field_name, field.description)
        field_info = FieldInfo(description=description, default=field.default)
        if field.metadata:
            field_info.metadata = field.metadata
        fields[field_name] = (field.annotation, field_info)

    rtn = create_model(  # type: ignore
        name, **fields, __config__=ConfigDict(arbitrary_types_allowed=True)
    )

    selected_annotations = [
        (name, annotation) for name, annotation in model.__annotations__.items() if name in field_names
    ]

    rtn.__annotations__ = dict(selected_annotations)
    rtn.__doc__ = textwrap.dedent(fn_description or model.__doc__ or "")
    return rtn


def _rm_titles(kv: dict, prev_key: str = "") -> dict:
    new_kv = {}
    for k, v in kv.items():
        if k == "title":
            if isinstance(v, dict) and prev_key == "properties" and "title" in v:
                new_kv[k] = _rm_titles(v, k)
            else:
                continue
        elif isinstance(v, dict):
            new_kv[k] = _rm_titles(v, k)
        else:
            new_kv[k] = v
    return new_kv
